type loop

type unloop_how = 
  | UnloopOne				(** unloop the innermost loop *)
  | UnloopAll				(** unloop all loops *)

type revent =
  | READ				(* ev_io detected read will not block *)
  | WRITE				(* ev_io detected write will not block *)
  | IOFDSET				(* internal use only *)
  | TIMEOUT				(* timer timed out *)
  | PERIODIC				(* periodic timer timed out *)
  | SIGNAL				(* signal was received *)
  | CHILD				(* child/pid had status change *)
  | STAT				(* stat data changed *)
  | IDLE				(* event loop is idling *)
  | PREPARE				(* event loop about to poll *)
  | CHECK				(* event loop finished poll *)
  | EMBED				(* embedded event loop needs sweep *)
  | FORK				(* event loop resumed in child *)
  | ASYNC                               (* an asynchronous signal was received *)
  | ERROR				(* sent when an error occurs *)

type revents = revent list

type io_opaque
type loop_opaque

type generic_io =
    {
      io_cb		: generic_cb;
      io_io		: io_opaque;
      mutable io_loop	: loop_opaque;
    }

and generic_cb = generic_io -> revents -> unit

type 'a ev = generic_io

exception NoLoop
exception LoopInitFailed
exception VersionMismatch
exception StaleObject

type io
type timer
type signal
type idle
type prepare
type check
type periodic
type child
type fork
type stat
type async

type reschedule_cb = periodic ev -> float -> float

type periodic_time =
  | Absolute of float
  | Repeating of float * float
  | Manual of reschedule_cb

type 'a cb = 'a ev -> revents -> unit
type loop_flag =
    [ `NOENV
    | `FORKCHECK
    ]

type run_loop_flag =
  | OneShot
  | NonBlock

type backend =
    [ `SELECT
    | `POLL
    | `EPOLL
    | `KQUEUE
    | `DEVPOLL
    | `PORT]

type loop_arg = [ loop_flag | backend ]

type exn_handler = loop -> exn -> unit

let string_of_loop_arg : [< loop_arg] -> string = function
  | `NOENV -> "NOENV"
  | `FORKCHECK -> "FORKCHECK"
  | `SELECT -> "SELECT"
  | `POLL -> "POLL"
  | `EPOLL -> "EPOLL"
  | `KQUEUE -> "KQUEUE"
  | `DEVPOLL -> "DEVPOLL"
  | `PORT -> "PORT"

let string_of_revent : revent -> string = function
  | READ -> "READ"
  | WRITE -> "WRITE"
  | IOFDSET -> "IOFDSET"
  | TIMEOUT -> "TIMEOUT"
  | PERIODIC -> "PERIODIC"
  | SIGNAL -> "SIGNAL"
  | CHILD -> "CHILD"
  | STAT -> "STAT"
  | IDLE -> "IDLE"
  | PREPARE -> "PREPARE"
  | CHECK -> "CHECK"
  | EMBED -> "EMBED"
  | FORK -> "FORK"
  | ASYNC -> "ASYNC"
  | ERROR -> "ERROR"

  
external embeddable_backends : unit -> backend list = "evo_embeddable_backends"
external supported_backends : unit -> backend list = "evo_supported_backends"
external recommended_backends : unit -> backend list = "evo_recommended_backends"

external loop_new : loop_arg list -> loop = "evo_loop_new"
external loop_destroy : loop -> unit = "evo_loop_destroy"
external default_loop : loop_arg list -> loop = "evo_default_loop"
external loop_now : loop -> float = "evo_loop_now"
external is_default_loop : loop -> bool = "evo_is_default_loop"
external loop_count : loop -> int = "evo_loop_count"
external set_exn_handler : loop -> exn_handler -> unit = "evo_set_exn_handler"

external io_init : Unix.file_descr -> revents -> io cb -> io ev = "evo_io_init"

external idle_init : idle cb -> idle ev = "evo_idle_init"

external async_init : async cb -> async ev = "evo_async_init"
external async_send : async ev -> unit = "evo_async_send"

external stat_init : string -> float -> stat cb -> stat ev = "evo_stat_init"
external stat_set : stat ev -> string -> float -> unit = "evo_stat_set"
external stat_get_attr : stat ev -> Unix.stats = "evo_stat_get_attr"
external stat_get_prev : stat ev -> Unix.stats = "evo_stat_get_attr"

external fork_init : idle cb -> idle ev = "evo_fork_init"

external prepare_init : prepare cb -> prepare ev = "evo_prepare_init"

external check_init : check cb -> check ev = "evo_check_init"

external timer_init : float -> float -> timer cb -> timer ev = "evo_timer_init"
external timer_again : timer ev -> unit = "evo_timer_again"

external periodic_init : periodic_time -> periodic cb -> periodic ev = "evo_periodic_init"
external periodic_set : periodic ev -> periodic_time -> unit = "evo_periodic_set"
external periodic_get_at : periodic ev -> float = "evo_periodic_get_at"

external signal_init : int -> signal cb -> signal ev = "evo_signal_init"

external child_init : int -> bool -> child cb -> child ev = "evo_child_init"
external child_set : child ev -> int -> bool -> unit = "evo_child_set"
external child_get_pid : child ev -> int = "evo_child_get_pid"
external child_get_rpid : child ev -> int = "evo_child_get_rpid"
external child_get_rstatus : child ev -> int = "evo_child_get_rstatus"

external loop : loop -> run_loop_flag list -> unit = "evo_loop"
external unloop : loop -> unloop_how -> unit = "evo_unloop"
external forked : loop -> unit = "evo_forked"
external set_io_collect_interval : loop -> float -> unit = "evo_set_io_collect_interval"
external set_timeout_collect_interval : loop -> float -> unit = "evo_set_timeout_collect_interval"

external set_cb : 'a ev -> 'a cb -> unit = "evo_set_cb"
external get_cb : 'a ev -> 'a cb = "evo_get_cb"
let no_cb : 'a cb = fun _ _ -> ()
external is_active : 'a ev -> bool = "evo_is_active"
external is_pending : 'a ev -> bool = "evo_is_pending"
external get_priority : 'a ev -> int = "evo_get_priority"
external set_priority : 'a ev -> int -> unit = "evo_get_priority"
external get_loop : 'a ev -> loop = "evo_get_loop"
external invoke : 'a ev -> revents -> unit = "evo_invoke"
external clear_pending : 'a ev -> revents = "evo_clear_pending"
external start : loop -> 'a ev -> unit = "evo_start"
external stop : 'a ev -> unit = "evo_stop"
external set_keepalive : 'a ev -> bool -> unit = "evo_set_keepalive"

let default_exn_handler _ exn =
  Printf.eprintf "ev: Ignoring leaking exception: %s\n%!" (Printexc.to_string exn);
  ()

