(** Ev is a module for binding
    {{:http://software.schmorp.de/pkg/libev.html}libev}, which is a
    high-performance event loop. You may wish to browse
    {{:http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod}libev's
    documentation} from time to time, but this documentation attempts to
    be as self-containing API documentation as possible, with links to
    relevant libev's documentation where appropriate.
*)

(** loop_flag is used when constructing a new loop with loop_new or default_loop 
    
    @see <VALloop_new> loop_new
    @see <VALdefault_loop> default_loop
    @see <http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#FUNCTIONS_CONTROLLING_THE_EVENT_LOOP>Functions controlling the event loop (ev_default_loop)
*)
type loop_flag =
    [ `NOENV				(** don't respect LIBEV_FLAGS environment variable *)
    | `FORKCHECK			(** automatically check for forking in the event loop *)
    ]

type run_loop_flag =
  | OneShot
  | NonBlock

(** supported backends 

    @see <http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#FUNCTIONS_CONTROLLING_THE_EVENT_LOOP>Functions controlling the event loop (ev_default_loop)
*)
type backend =
    [ `SELECT
    | `POLL
    | `EPOLL
    | `KQUEUE
    | `DEVPOLL
    | `PORT]

(** loop_new and default_loop construction argument 
    
    @see <VALloop_new> loop_new
    @see <VALdefault_loop> default_loop
*)
type loop_arg = [ loop_flag | backend ]

(** how to stop a loop

    @see <VALloop> loop*)
type unloop_how = 
  | UnloopOne				(** unloop the innermost loop *)
  | UnloopAll				(** unloop all loops *)

(** event types that are recognized 

    @see <http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#ANATOMY_OF_A_WATCHER> Anatomy of a watcher
*)
type revent =
  | READ				(** ev_io detected read will not block *)
  | WRITE				(** ev_io detected write will not block *)
  | IOFDSET				(** internal use only *)
  | TIMEOUT				(** timer timed out *)
  | PERIODIC				(** periodic timer timed out *)
  | SIGNAL				(** signal was received *)
  | CHILD				(** child/pid had status change *)
  | STAT				(** stat data changed *)
  | IDLE				(** event loop is idling *)
  | PREPARE				(** event loop about to poll *)
  | CHECK				(** event loop finished poll *)
  | EMBED				(** embedded event loop needs sweep *)
  | FORK				(** event loop resumed in child *)
  | ASYNC                               (** an asynchronous signal *)
  | ERROR				(** sent when an error occurs *)

exception NoLoop			(** no loop started for an event when one required *)
exception LoopInitFailed		(** failed to create a loop object *)
exception VersionMismatch		(** there is too great version difference to use system libev *)
exception StaleObject			(** an attempt to access a destroyed object was made *)

type revents = revent list		(** list of events received *)

type loop				(** abstract data type for event loops *)

type 'a ev				(** abstract data type for events *)

type 'a cb = 'a ev -> revents -> unit	(** callback function; called when an event occurs *)

type exn_handler = loop -> exn -> unit  (** an exception handler; if this function throws, it will end
					    the event loop which will forward the exception *)

(** convert a loop argument into a human-readable string *)
val string_of_loop_arg : [< loop_arg] -> string

(** convert an revent into a human-readable string *)
val string_of_revent : revent -> string

(** {2 Backend information} *)

(** list of backends that may be embedded; embedding not yet supported 
    
    @see <http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#GLOBAL_FUNCTIONS> libev global functions (ev_embeddable_backends)
*)
external embeddable_backends : unit -> backend list = "evo_embeddable_backends"

(** list of supported backends 
    
    @see <http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#GLOBAL_FUNCTIONS> libev global functions (ev_supported_backends)
*)
external supported_backends : unit -> backend list = "evo_supported_backends"

(** list of recommended backends for this host 
    
    @see <http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#GLOBAL_FUNCTIONS> libev global functions (ev_recommended_backends)
*)
external recommended_backends : unit -> backend list = "evo_recommended_backends"

(** {2 Event loop management} *)

(** access the default loop. the next call will ignore the flags and return the same loop. 
    note: do NOT destroy this loop if you wish to use it again, calling default_loop
    again won't bring it back. (libev bug?)
    
    @raise LoopInitFailed if the initialization failed
    @raise VersionMismatch if the system library is too old compared to the library this wrapper was compiled against
*)
external default_loop : loop_arg list -> loop = "evo_default_loop"

(** create a new event loop 
    
    @raise LoopInitFailed if the initialization of the loop failed
    @raise VersionMismatch if the system library is too old compared to the library this wrapper was compiled against
    @see <http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#FUNCTIONS_CONTROLLING_THE_EVENT_LOOP>Functions controlling the event loop (ev_default_loop)
*)
external loop_new : loop_arg list -> loop = "evo_loop_new"

(** destroy an event loop. Future access to the object will raise
    StaleObject. note: do NOT destroy the default loop if you wish to use
    it again, calling default_loop again won't bring it back. (libev bug?)
*)
external loop_destroy : loop -> unit = "evo_loop_destroy"

(** time of day according to the event loop 

    @see <http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#FUNCTIONS_CONTROLLING_THE_EVENT_LOOP>Functions controlling the event loop (ev_loop_new)
*)
external loop_now : loop -> float = "evo_loop_now"

(** is this the default loop? *)
external is_default_loop : loop -> bool = "evo_is_default_loop"

(** a monotonic (wrapping) counter that roughly indicates the number of times the
    event loop has iterated

    @see <http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#FUNCTIONS_CONTROLLING_THE_EVENT_LOOP>Functions controlling the event loop (ev_loop_count)
*)
external loop_count : loop -> int = "evo_loop_count"

(** loop l flags - run an event loop
    
    @param l the event loop object
    @param flags loop flags; may be ignored
*)
external loop : loop -> run_loop_flag list -> unit = "evo_loop"

(** unloop l how - run an event loop. Running loop will reset the unloop state of the
    event loop object.
    
    @param l the event loop object
    @param how how to unloop: UnloopOne or UnloopAll
*)
external unloop : loop -> unloop_how -> unit = "evo_unloop"

(** set_exn_handler loop handler - set the exception handler of an event loop
    
    @param loop the event loop object
    @param handler the event handler
*)
external set_exn_handler : loop -> exn_handler -> unit = "evo_set_exn_handler"

(** default exception handler; it output diagnostics to stderr and swallows the exception
*)
val default_exn_handler : loop -> exn -> unit

(** forked loop - inform an event loop that we have forked (and we'd like to keep using the libev services)

    @param loop the event loop object
*)
external forked : loop -> unit = "evo_forked"

(** set_io_collect_interval loop interval - collect io events in
    lumps. may help throughput in high load situations by avoiding work,
    but will increase latency.
*)
external set_io_collect_interval : loop -> float -> unit = "evo_set_io_collect_interval"

(** set_timeout_collect_interval loop interval - collect timeout events in
    lumps. may help throughput in high load situations by avoiding work,
    but will increase latency.
*)
external set_timeout_collect_interval : loop -> float -> unit = "evo_set_timeout_collect_interval"


(** {2 Event object management} *)

(** set_cb ev f - sets the callback of an event 

    @param ev the event object to change
    @param f the new function
*)
external set_cb : 'a ev -> 'a cb -> unit = "evo_set_cb"

(** get_cb ev - retrieves the callback of an event

    @param ev the event object to query
    @return the callback function
*)
external get_cb : 'a ev -> 'a cb = "evo_get_cb"

(** no_cb ev - a callback that does nothing; could be useful
    as the callback parameter to _init functions.

    @return the callback function
*)
val no_cb : 'a cb

(** is_active ev - is this event active? an event is
    active when it has been issued *_start an inactive
    after a *_stop.

    @param ev the event object to query
    @return the status
*)
external is_active : 'a ev -> bool = "evo_is_active"

(** is_pending ev - does this even have outstanding events?

    @param ev the event object to query
    @return the status
*)
external is_pending : 'a ev -> bool = "evo_is_pending"

(** get_priority ev - retrieve the priority of an event. 
    values range from -2 to 2.

    @param ev the event object to query
    @return the status
*)
external get_priority : 'a ev -> int = "evo_get_priority"

(** set_priority ev priority - set the priority of an event. 
    Priority should range from -2 to 2 and it may of may not
    be clamped by libev. (??)

    @param ev the event object to modify
*)
external set_priority : 'a ev -> int -> unit = "evo_get_priority"

(** keepalive ev state - set the the significance of this event
    watcher from its loop, when considering whether to exit or not. If the
    state is set to true and this is the last event stopping the event
    loop from exiting, it will exit.

    @param ev the event object to modify
    @param state the new state
*)
external set_keepalive : 'a ev -> bool -> unit = "evo_set_keepalive"

(** raises NoLoop the event is not active 

    @see <#VALis_active> is_active
    @raise NoLoop if the event is not active
*)
external get_loop : 'a ev -> loop = "evo_get_loop"

(** invoke an event as if it had just been fired. the
    event must be registered to a loop, or NoLoop is thrown. 

    @raise NoLoop if the event is not active
*)
external invoke : 'a ev -> revents -> unit = "evo_invoke"

(** clear pending events from an event watcher: return the 
    list if events that were pending *)
external clear_pending : 'a ev -> revents = "evo_clear_pending"

(**
   start loop ev - attach an event watcher to an event loop (that is, activate it).
   If the event watcher is already active (with the same loop), do nothing.

   @param loop the event loop to attach this event to
   @param ev the event watcher
   @raise Invalid_argument the event loop is started to another loop
*)
external start : loop -> 'a ev -> unit = "evo_start"

(**
   stop ev - detach an event watcher from an event loop (thus deactivating it).
   If the event watcher is already inactive, do nothing.

   @param ev the event watcher
*)
external stop : 'a ev -> unit = "evo_stop"

(**************************************************************************)
(** {2 Event-specific functions} *)

(** {3 Io events} 
    A file descriptor will not block when read from or written to.

    Libev documentation: {{:http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#code_ev_io_code_is_this_file_descrip}Libev ev_io}
*)

type io					(** abstract data type containing the event watcher *)

(** io_init f fd revents 

    @param f function to call, of type io cb -> revents -> unit. Use no_cb if you don't want to define any now.
    @param fd file descriptor to watch
    @param revents events to watch: READ of WRITE
*)
external io_init : Unix.file_descr -> revents -> io cb -> io ev = "evo_io_init"

(**************************************************************************)
(** {3 Timer events} 
    Fired after a certain time period; optionally recurringly.

    Libev documention: {{:http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#code_ev_timer_code_relative_and_opti}Libev ev_timer}
*)

type timer				(** abstract data type containing the event watcher *)

(** timer_init f offset repeat - call a callback after offset seconds;
    if repeat > 0.0, retrigger the timer to run periodically.

    @param f function to call, of type timer cb -> revents -> unit. Use no_cb if you don't want to define any now.
*)
external timer_init : float -> float -> timer cb -> timer ev = "evo_timer_init"

(** timer_again ev - retrigger the timer to run again after the 
    interval.

    @param ev the timer event
    @raise NoLoop if this event is not active
    @see <#VALtimer_init> timer_init
    @see <#VALis_active> is_active
*)
external timer_again : timer ev -> unit = "evo_timer_again"

(**************************************************************************)
(** {3 Periodic events} 
    Fired with a certain periodic schedule.

    Libev documention: {{:http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#code_ev_periodic_code_to_cron_or_not> ev_periodic}Libev ev_periodic}
*)

type periodic				(** abstract data type containing the event watcher *)

(** callback for rescheduling next event, for periodic events 

    @see <VALperiodic_init> periodic_init
    @see <http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#Watcher_Specific_Functions_and_Data_-3>ev_periodic_init
*)
type reschedule_cb = periodic ev -> float -> float

(** describes the desired times when a periodic event should fire 

    @see <VALperiodic_init> periodic_init
    @see <http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#Watcher_Specific_Functions_and_Data_-3>ev_periodic_init
*)
type periodic_time =
  | Absolute of float			(** absolute unix time *)
  | Repeating of float * float		(** offset * interval - repeat every interval seconds, with unix time offset offset *)
  | Manual of reschedule_cb		(** use a custom rescheduling function *)


(** periodic_init f periodic_time - call a callback with a certain
    schedule.
    
    @param f function to call, of type periodic cb -> revents -> unit. Use no_cb if you don't want to define any now.
    @return the periodic event object, still detached from event loop
    @see <http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#code_ev_periodic_code_to_cron_or_not> ev_periodic
*)
external periodic_init : periodic_time -> periodic cb -> periodic ev = "evo_periodic_init"

(** periodic_set ev periodic_time - set the periodic parameters again

    @see <#VALperiodic_init> periodic_init
*)
external periodic_set : periodic ev -> periodic_time -> unit = "evo_periodic_set"

(** periodic_get_at ev - retrieve the epoch time when the event
    will be fired the next time

    @param ev the periodic event
*)
external periodic_get_at : periodic ev -> float = "evo_periodic_get_at"

(**************************************************************************)
(** {3 Unix signals} 
    Fired when a unix signal is received.

    Libev documention: {{:http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#code_ev_signal_code_signal_me_when_a}Libev ev_signal}
*)

type signal				(** abstract data type containing the event watcher *)

(** signal_init f signal - call a callback when a certain signal
    is raised.
    
    @param f function to call, of type signal cb -> revents -> unit. Use no_cb if you don't want to define any now.
    @return the signal event object, still detached from event loop
*)
external signal_init : int -> signal cb -> signal ev = "evo_signal_init"

(**************************************************************************)
(** {3 Asynchronous signals} 
    Fired when a async_send is called.

    Libev documention: {{:http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#code_ev_async_code_how_to_wake_up_an}Libev ev_async}
*)

type async

(** async_init f - call a callback when a async_send is called

    @param f the function to call
*)
external async_init : async cb -> async ev = "evo_async_init"

(** async_send ev - issue an asynchronous event *)
external async_send : async ev -> unit = "evo_async_send"

(**************************************************************************)
(** {3 Child process events} 
    Fired when the status of a child process changes.

    Libev documention: {{:http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#code_ev_child_code_watch_out_for_pro> Libev ev_timer}Libev ev_child}
*)

type child				(** abstract data type containing the event watcher *)

(** child_init f child trace - call a callback when the status of a child
    process changes. 
    
    @param f function to call, of type child cb -> revents -> unit. Use no_cb if you don't want to define any now.
    @param trace if true, activate also when the process is stopped
    or continued, not only when it exits
    @return the child event object, still detached from event loop
*)
external child_init : int -> bool -> child cb -> child ev = "evo_child_init"

(** child_set ev child trace - set the child event parameters again

    @see <#VALchild_init> child_init
*)
external child_set : child ev -> int -> bool -> unit = "evo_child_set"

(** child_get_pid ev - retrieve the pid that is being watched

    @param ev the child event
    @return the monitored process pid
*)
external child_get_pid : child ev -> int = "evo_child_get_pid"

(** child_get_rpid ev - retrieve the pid that issued a status change

    @param ev the child event
    @return the child process pid
*)
external child_get_rpid : child ev -> int = "evo_child_get_rpid"

(** child_get_status ev - retrieve the status of the child process

    @param ev the child event
    @return the child process status
*)
external child_get_rstatus : child ev -> int = "evo_child_get_rstatus"

(**************************************************************************)
(** {3 Stat events} 
    Fired when the state of a file changes.

    Libev documentation: {{:http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#code_ev_stat_code_did_the_file_attri}Libev ev_stat}
*)

type stat				(** abstract data type containing the event watcher *)

(** stat_init f filename interval - call a callback when the state of a file (as seen by stat) changes

    @param f function to call, of type stat cb -> revents -> unit. Use no_cb if you don't want to define any now.
*)
external stat_init : string -> float -> stat cb -> stat ev = "evo_stat_init"

(** stat_set filename interval - sets the parameters of the event watcher.
*)
external stat_set : stat ev -> string -> float -> unit = "evo_stat_set"

(** stat_get_attr ev - retrieve the stats of the monitored file
*)
external stat_get_attr : stat ev -> Unix.stats = "evo_stat_get_attr"

(** stat_get_attr ev - retrieve the old stats of the monitored file (befoe the callback)
*)
external stat_get_prev : stat ev -> Unix.stats = "evo_stat_get_attr"

(**************************************************************************)
(** {3 Idle events} 
    Fired when the event loop has nothing else to do. 

    Libev documentation: {{:http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#code_ev_idle_code_when_you_ve_got_no}Libev ev_idle}
*)

type idle				(** abstract data type containing the event watcher *)

(** idle_init f - call a callback when the event loop has nothing else to do 

    @param f function to call, of type idle cb -> revents -> unit. Use no_cb if you don't want to define any now.
*)
external idle_init : idle cb -> idle ev = "evo_idle_init"

(**************************************************************************)
(** {3 Fork events} 
    Fired after a fork has been done (and the event loop notices it/is notified about it).

    Libev documentation: {{:http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#code_ev_fork_code_the_audacity_to_re}Libev ev_fork}
*)

type fork				(** abstract data type containing the event watcher *)

(** fork_init f - call a callback after a fork (and the event loop notices it/is notified about it).

    @param f function to call, of type fork cb -> revents -> unit. Use no_cb if you don't want to define any now.
*)
external fork_init : fork cb -> fork ev = "evo_fork_init"

(**************************************************************************)
(** {3 Prepare/check events} 
    Fired before and after waiting for events.

    Libev documention: {{:http://pod.tst.eu/http://cvs.schmorp.de/libev/ev.pod#code_ev_prepare_code_and_code_ev_che}Libev ev_prepare/check}
*)

type prepare				(** abstract data type containing the event watcher *)
type check				(** abstract data type containing the event watcher *)

(** prepare_init f - call a callback when the event loop is about to start
    waiting for events

    @param f function to call, of type prepare cb -> revents -> unit. Use no_cb if you don't want to define any now.
*)
external prepare_init : prepare cb -> prepare ev = "evo_prepare_init"

(** check_init f - call a callback when the event loop is about to start
    calling the callbacks after receiving the events

    @param f function to call, of type check cb -> revents -> unit. Use no_cb if you don't want to define any now.
*)
external check_init : check cb -> check ev = "evo_check_init"

(** {1 Examples} 

{2 Complete example: receive lines with timeout}

A complete example of semi-efficiently receiving data from stdin and 
timing out when there has not been input for a short interval (5 seconds).

For some reason this example can misbehave when copy-pasted to a toplevel.
#use works fine, though. The problem is that the timeout is fired either
instantaneously or otherwise too fast (in about 3 seconds).

{[
(* function for feeding strings to buffer and retrieving whole lines *)
let add_buf_str buf str length =
  let rec add_buf_at buf ofs = 
    let rec find_eol at =
      if at < length then
        if str.[at] = '\n' then
          Some at
        else
          find_eol (at + 1)
       else
         None
    in
      match find_eol ofs with
       | None -> 
         if length <> ofs then
           Buffer.add_substring buf str ofs (length - ofs);
         []
       | Some at ->
         Buffer.add_substring buf str ofs (at - ofs);
         let str = Buffer.contents buf in
           Buffer.clear buf;
           str::add_buf_at buf (at + 1)
  in
    add_buf_at buf 0

let main () = 
  (* the default event loop *)
  let loop = Ev.default_loop [] in

  (* intermediate storage for incoming bytes *)
  let buf = Buffer.create 1024 in

  let exit_io io =
    Printf.printf "Bye!\n%!"; 
    Ev.stop io
  in

  (* a receiving buffer for a single message *)
  let str_buf = String.create 1024 in
  (* start io immediately after creating it *)
  let io_wait = Ev.io_init Unix.stdin [Ev.READ] Ev.no_cb in
  let timeout_wait = Ev.timer_init 5.0 5.0 Ev.no_cb in
    Ev.set_cb io_wait (
      (* use io parameter for an easy mechanism to stop itself *)
      (* revents parameter is ignored *)
      fun io _ ->
        (* receive data from the peer.. (stdin in this case) *)
        try
          let n = Unix.read Unix.stdin str_buf 0 (String.length str_buf) in
            (* bail out when there is an EOF or an error (doesn't handle EAGAIN) *)
            if n <= 0 then exit_io io
            else
              ((* reset timer *)
                Ev.timer_again timeout_wait;
                (* output received messages *)
                List.iter
                  (fun str -> 
                     Printf.printf "Received: %s\n%!" str;
                     if str = "bye" then
		       begin
			 Ev.stop io;
			 Ev.stop timeout_wait;
		       end
                  )
                  (add_buf_str buf str_buf n)
              )
        with 
          | Unix.Unix_error ((Unix.EAGAIN | Unix.EINTR), _, _) -> ()
          | Unix.Unix_error _ -> exit_io io
    );
    Ev.set_cb timeout_wait (
      fun io revents ->
        Printf.printf "Timeout! %s\n%!" (String.concat " " (List.map Ev.string_of_revent revents));
	Ev.unloop (Ev.get_loop io) Ev.UnloopOne;
    );

    Ev.start loop io_wait;
    Ev.start loop timeout_wait;

    (* set the descriptor to a non-blocking mode just in case (certain platforms may produce
       spurious events) *)
    Unix.set_nonblock Unix.stdin;
    (* start it! *)
    Ev.loop loop [];
    Unix.clear_nonblock Unix.stdin

let () = main ()
]}

*)



