(** This module provides some callbacks to the C module, written
    in Ocaml for convenience *)

open Ev

let rec mask_of_loop_flags flags =
  match flags with
    | [] -> 0
    | hd::tl ->
	mask_of_loop_flags tl lor
	  match hd with
	    | `NOENV -> 0x01000000 (* do NOT consult environment *)
	    | `FORKCHECK -> 0x02000000 (* check for a fork in each iteration *)
	    | `SELECT  -> 0x00000001 (* about anywhere *)
	    | `POLL    -> 0x00000002 (* !win *)
	    | `EPOLL   -> 0x00000004 (* linux *)
	    | `KQUEUE  -> 0x00000008 (* bsd *)
	    | `DEVPOLL -> 0x00000010 (* solaris 8 *) (* NYI *)
	    | `PORT    -> 0x00000020 (* solaris 10 *)

let rec loop_flags_of_mask flags =
  if flags land 0x01000000 <> 0 then [`NOENV] else [] @
  if flags land 0x02000000 <> 0 then [`FORKCHECK] else [] @
  if flags land 0x00000001 <> 0 then [`SELECT] else [] @
  if flags land 0x00000002 <> 0 then [`POLL] else [] @
  if flags land 0x00000004 <> 0 then [`EPOLL] else [] @
  if flags land 0x00000008 <> 0 then [`KQUEUE] else [] @
  if flags land 0x00000010 <> 0 then [`DEVPOLL] else [] @
  if flags land 0x00000020 <> 0 then [`PORT] else []

let rec mask_of_run_loop_flags flags =
  match flags with
    | [] -> 0
    | hd::tl ->
	mask_of_run_loop_flags tl lor
	  match hd with
	    | OneShot -> 0x02 (* block *once* only *)
	    | NonBlock -> 0x01  (* do not block/wait *)

let rec mask_of_revents flags =
  match flags with
    | [] -> 0l
    | hd::tl ->
	Int32.logor (mask_of_revents tl) 
	  (match hd with
	     | READ           -> 0x01l (* ev_io detected read will not block *)
	     | WRITE          -> 0x02l (* ev_io detected write will not block *)
	     | IOFDSET        -> 0x80l (* internal use only *)
	     | TIMEOUT  -> 0x00000100l (* timer timed out *)
	     | PERIODIC -> 0x00000200l (* periodic timer timed out *)
	     | SIGNAL   -> 0x00000400l (* signal was received *)
	     | CHILD    -> 0x00000800l (* child/pid had status change *)
	     | STAT     -> 0x00001000l (* stat data changed *)
	     | IDLE     -> 0x00002000l (* event loop is idling *)
	     | PREPARE  -> 0x00004000l (* event loop about to poll *)
	     | CHECK    -> 0x00008000l (* event loop finished poll *)
	     | EMBED    -> 0x00010000l (* embedded event loop needs sweep *)
	     | FORK     -> 0x00020000l (* event loop resumed in child *)
	     | ASYNC    -> 0x00040000l (* an asynchronous signal was received *)
	     | ERROR    -> 0x80000000l) (* sent when an error occurs *)

let revent_table =
  [|READ, 0x01l; (* ev_io detected read will not block *)
    WRITE, 0x02l; (* ev_io detected write will not block *)
    IOFDSET, 0x80l; (* internal use only *)
    TIMEOUT, 0x00000100l; (* timer timed out *)
    PERIODIC, 0x00000200l; (* periodic timer timed out *)
    SIGNAL, 0x00000400l; (* signal was received *)
    CHILD, 0x00000800l; (* child/pid had status change *)
    STAT, 0x00001000l; (* stat data changed *)
    IDLE, 0x00002000l; (* event loop is idling *)
    PREPARE, 0x00004000l; (* event loop about to poll *)
    CHECK, 0x00008000l; (* event loop finished poll *)
    EMBED, 0x00010000l; (* embedded event loop needs sweep *)
    FORK, 0x00020000l; (* event loop resumed in child *)
    ASYNC, 0x00040000l; (* an asynchronous signal was received *)
    ERROR, 0x80000000l; (* sent when an error occurs *)
  |]

let rec revents_of_mask flags =
  Array.fold_left
    (fun a (symbol, value) ->
       if Int32.logand flags value <> 0l then
	 symbol::a
       else 
	 a
    )
    []
    revent_table

let () =
  Callback.register "mask_of_loop_flags" mask_of_loop_flags;
  Callback.register "loop_flags_of_mask" loop_flags_of_mask;
  Callback.register "mask_of_revents" mask_of_revents;
  Callback.register "revents_of_mask" revents_of_mask;
  Callback.register "handle_watcher_exception" default_exn_handler;
  Callback.register "mask_of_run_loop_flags" mask_of_run_loop_flags;
  Callback.register_exception "no_loop_exception" NoLoop;
  Callback.register_exception "loop_init_failed_exception" LoopInitFailed;
  Callback.register_exception "version_mismatch_exception" VersionMismatch;
  Callback.register_exception "stale_object_exception" StaleObject;
  ()
