open OUnit

let tests = ref []

let register label test =
  tests := (label >:: test)::!tests

let run f = f ()

let ( @@ ) a b = a b

let with_loop f = 
  bracket
    (fun () -> Ev.loop_new [])
    f
    Ev.loop_destroy

let with_default f = 
  bracket
    (fun () -> 
       let l = Ev.default_loop [] in
	 Ev.set_exn_handler l (fun loop exn -> raise exn);
	 l)
    f
    (fun _ -> ())
;;

register "new/destroy" @@
  fun () ->
    for c = 0 to 10000 do
      let l = Ev.loop_new [] in
	Ev.loop_destroy l
    done
;;

register "empty loop" @@
  with_loop Ev.loop
;;

register "timer loop" @@
  fun () ->
    let before = Unix.gettimeofday () in
      with_default
	(fun loop ->
	   let timer = Ev.timer_init 1.0 0.0 (fun _ _ -> Printf.printf "Timeout\n%!") in
	     Ev.start loop timer;
	     Ev.loop loop []
	) ();
      let after = Unix.gettimeofday () in
      let delta = after -. before in
	assert_bool "time didn't go forward" (after > before);
	assert_bool "timer didn't take more than 0.9 seconds" (delta >= 0.9);
	assert_bool "timer didn't take less than 1.1 seconds" (delta <= 1.1);
;;

register "destroy loop while looping.." @@
  fun () ->
    run @@
      bracket
      (fun () -> Ev.loop_new [])
      (fun loop ->
	 let timer = Ev.timer_init 0.1 0.0 (fun io _ -> Ev.stop io; Ev.loop_destroy loop) in
	   Ev.start loop timer;
	   Ev.loop loop []
      )
      Ev.loop_destroy
;;

register "throw while handling" @@
  fun () ->
    run @@
      bracket
      (fun () -> Ev.loop_new [])
      (fun loop ->
	 let timer = Ev.timer_init 0.1 0.0 (fun io _ -> failwith "argh") in
	   Ev.start loop timer;
	   Ev.loop loop []
      )
      Ev.loop_destroy
;;

register "throw while handling, with special handler" @@
  fun () ->
    for c = 0 to 10 do 
      let caught = ref 0 in
	run @@
	  bracket
	  (fun () -> 
	     let l = Ev.loop_new [] in
	       Ev.set_exn_handler l (fun loop exn -> incr caught);
	       l
	  )
	  (fun loop ->
	     let timer = Ev.timer_init 0.1 0.0 (fun io _ -> failwith "argh") in
	       Ev.start loop timer;
	       Ev.loop loop []
	  )
	  Ev.loop_destroy;
	assert_bool "expected exactly 1 caught exception" (!caught = 1);
    done
;;

register "throw while handling, with forwarding handler" @@
  fun () ->
    for c = 0 to 10 do 
      let module M = struct exception Foo end in
	assert_raises M.Foo @@
	  bracket
	  (fun () -> 
	     let l = Ev.loop_new [] in
	       Ev.set_exn_handler l (fun loop exn -> raise exn);
	       l
	  )
	  (fun loop ->
	     let timer = Ev.timer_init 0.01 0.0 (fun io _ -> raise (M.Foo)) in
	       Ev.start loop timer;
	       Ev.loop loop []
	  )
	  Ev.loop_destroy
    done
;;

register "send, receive" @@
  fun () ->
    for c = 0 to 10 do 
      let (fd'in, fd'out)  = Unix.pipe () in
      let received = ref 0 in
	with_default (
	  fun loop ->
	    Ev.start loop (Ev.io_init fd'in [Ev.READ] (fun io _ -> Ev.stop io; incr received));
	    Ev.start loop (Ev.timer_init 0.01 0.0 (fun t _ -> Ev.stop t; ignore (Unix.write fd'out "hello" 0 5)));
	    Ev.loop loop []
	) ();
	assert_bool "Didn't received at most 1 events" (!received <= 1);
	assert_bool "Didn't received at least 1 events" (!received >= 1);
    done
;;

register "multiple starts, stops" @@
  fun () ->
    for c = 0 to 10 do
      with_default (
	fun loop ->
	  let t = Ev.timer_init 0.1 0.0 (fun io _ -> Ev.stop io) in
	    Ev.start loop t;
	    Ev.start loop t;
	    Ev.stop t;
	    Ev.stop t;
	    Ev.start loop t;
	    Ev.start loop t;
	    Ev.loop loop []
      ) ()
    done
;;

register "start in two loops" @@
  fun () ->
    for c = 0 to 10 do
      assert_raises (Invalid_argument "Ev.ev_start: the event watcher is already started on another loop") @@
	with_default (
	  fun loop1 ->
	    with_loop (
	      fun loop2 ->
		let t = Ev.timer_init 0.1 0.0 (fun io _ -> Ev.stop io) in
		  Ev.start loop1 t;
		  Ev.start loop2 t
	    ) ()
	)
    done
;;


let main () =
  ignore (
    List.map (
      run_test_tt ~verbose:true
    ) (List.rev !tests)
  )

let () = main ()
